package ru.amse.baltijsky.javascheme.exporter.image;

import ru.amse.baltijsky.javascheme.nodeshape.*;
import ru.amse.baltijsky.javascheme.tree.*;
import ru.amse.baltijsky.javascheme.tree.walker.TraversalState.Rel;
import ru.amse.baltijsky.javascheme.util.Dir;
import ru.amse.baltijsky.javascheme.util.Size;
import static ru.amse.baltijsky.javascheme.util.Util.max;
import static ru.amse.baltijsky.javascheme.util.VisConst.XGAP;
import static ru.amse.baltijsky.javascheme.util.VisConst.YGAP;

import java.awt.*;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Builds a tree that corresponds to the schema tree but contains only the information about how the nodes should be drawn.
 * The resulting tree depends on the graphics context, namely, the settings for the font etc, so a graphic context is needed.
 */
public class VisualTreeBuilder {
    /**
     * Direction shorthands.
     */
    private static final int L = Dir.l();
    private static final int R = Dir.r();
    private static final int U = Dir.u();
    private static final int D = Dir.d();

    /**
     * A one-way implementation of the tree builder.
     */
    private class VisualTreeBuilderImpl {
        /**
         * The schema tree upon which to create visual tree.
         */
        private final SchemaTree schemaTree;
        /**
         * The graphics context for which the visual tree is built.
         */
        private final Graphics graphics;

        /**
         * Creates a new visual tree builder implementation for the specified schema tree.
         *
         * @param theSchemaTree the schema tree
         * @param g             the graphics context for building the visual tree
         */
        public VisualTreeBuilderImpl(SchemaTree theSchemaTree, Graphics g) {
            schemaTree = theSchemaTree;
            graphics = g;
        }

        // implementation of the outer class method
        public VisualNode build() {
            // create a special fake schema node that has the shape of the initial state
            // its next node is the root of the original schema tree
            SchemaNode startSchemaNode = new SchemaNode() {
                public SchemaNode getNext() {
                    return schemaTree.getRoot();
                }

                public boolean hasNext() {
                    return schemaTree.getRoot() != null;
                }

                public SchemaNode getParent() {
                    return null;
                }

                public boolean hasParent() {
                    return false;
                }

                public Iterator<SchemaNode> getChildren() {
                    return new ArrayList<SchemaNode>().iterator();
                }

                public boolean hasChildren() {
                    return false;
                }

                public NodeType getNodeType() {
                    return NodeType.METHOD;
                }

                public INodeShape getShape() {
                    return new InitialShape();
                }
            };
            return visit(startSchemaNode, null, Rel.ROOT);
        }

        /**
         * Visits the specified schema node and creates a new visual node that has the specified relationship
         * with the given visual node.
         *
         * @param curSchemaNode  the schema node to visit
         * @param prevVisualNode the previously created node in the schema tree
         * @param rel            the relationship between the newly created node and the previous visual node
         * @return the created visual node that corresponds to the visited schema node
         */
        private VisualNode visit(SchemaNode curSchemaNode, VisualNode prevVisualNode, Rel rel) {
            if (curSchemaNode instanceof ModuleNode) {
                return skipModuleNodes(curSchemaNode, prevVisualNode, rel);
            }

            VisualNode node;
            // create? a new visual node with the proper shape and puts it in the place in the visual tree
            if (rel == Rel.NEXT) {
                node = new VisualNode(prevVisualNode.getParent(), prevVisualNode, curSchemaNode);
            } else if (rel == Rel.CHILD) {
                node = new VisualNode(prevVisualNode, null, curSchemaNode);
            } else { // if (rel == Rel.ROOT) {
                node = new VisualNode(null, null, curSchemaNode);
            }

            visitNext(node, curSchemaNode);

            return node;
        }

        /**
         * Visits the nodes reachable from the specified visual node that corresponds to the specified schema node,
         * calculating the sizes of this node's subtree.
         *
         * @param node          the visual node to visit its children and next nodes
         * @param curSchemaNode the corresponding schema node
         */
        private void visitNext(VisualNode node, SchemaNode curSchemaNode) {
            // we'll start with the size of the node and calculate the childrenSubtreeSize of the whole subtree with the root in this node
            Size childrenSubtreeSize = new Size();
            int[] nodeDim = node.getShape().getShapeDims(graphics);
            int[] closeNodeDim = node.getClosingShape().getShapeDims(graphics);
            int leftSubtreeSize = XGAP + max(nodeDim[L], closeNodeDim[L]);
            int rightSubtreeSize = max(nodeDim[R], closeNodeDim[R]) + XGAP;
            childrenSubtreeSize.x = leftSubtreeSize + rightSubtreeSize;
            childrenSubtreeSize.y = YGAP + nodeDim[U] + nodeDim[D] + YGAP;
            Size offset = new Size(leftSubtreeSize - nodeDim[L], YGAP);

            // for the children, we'll create special visual nodes without any accordance to the schema tree nodes
            // with the shape of the right-angle edge between the child and the parent, in some cases with label text.
            createChildren(curSchemaNode, node, nodeDim, offset, childrenSubtreeSize);
            node.setChildrenSubtreeSize(childrenSubtreeSize);
            rightSubtreeSize = childrenSubtreeSize.x - leftSubtreeSize;
            node.setOffset(offset);

            // calculate the size of the whole subtree for node
            int subtreeSizeY = childrenSubtreeSize.y;
            if (curSchemaNode.hasNext()) {
                VisualNode next = visit(curSchemaNode.getNext(), node, Rel.NEXT);
                int delta;
                if ((delta = next.getLeftSubtreeSize() - leftSubtreeSize) > 0) {
                    leftSubtreeSize += delta;
                }
                if ((delta = next.getRightSubtreeSize() - rightSubtreeSize) > 0) {
                    rightSubtreeSize += delta;
                }
                subtreeSizeY += next.getSubtreeSize().y;
            } else if (!node.hasParent()) {
                // for the last node in the outer level of the scheme add the final shape
                subtreeSizeY += addFinalShape(node);
            }
            if (node.hasChildren()) {
                subtreeSizeY += node.getClosingShape().getShapeDims(graphics)[D];
                subtreeSizeY += YGAP;
            }
            node.setSubtreeSize(leftSubtreeSize, rightSubtreeSize, subtreeSizeY);
        }

        /**
         * Adds the final shape to the visual tree (the next shape node for the node that corresponds to the last node
         * in SchemaTree in the outer level).
         *
         * @param node node that corresponds to the last node in SchemaTree in the outer level
         * @return the height of the subtree of the added node
         */
        private int addFinalShape(VisualNode node) {
            // create the node
            VisualNode finalNode = new VisualNode(node.getParent(), node, new FinalShape());
            // set the size
            Size finalNodeSize = finalNode.getShape().getBoundRectSize(graphics);
            finalNode.setChildrenSubtreeSize(new Size(XGAP + finalNodeSize.x + XGAP, YGAP + finalNodeSize.y + YGAP));
            int leftSize = finalNode.getShape().getShapeDims(graphics)[L] + XGAP;
            int width = finalNode.getChildrenSubtreeSize().x - leftSize;
            int height = finalNode.getChildrenSubtreeSize().y + YGAP;
            finalNode.setSubtreeSize(leftSize, width, height);

            return finalNode.getSubtreeSize().y;
        }

        /**
         * Skips over the module nodes that are too general for the schema of the method to be visualized.
         *
         * @param curSchemaNode  the module node
         * @param prevVisualNode previously created visual node
         * @param rel            the relatoinship between the previously created visual node and this node
         * @return the first node that is not a module node or (if no such exist) a no-shape emtpty node.
         */
        private VisualNode skipModuleNodes(SchemaNode curSchemaNode, VisualNode prevVisualNode, Rel rel) {
            if (curSchemaNode.hasChildren()) {
                return visit(curSchemaNode.getChildren().next(), prevVisualNode, rel);
            } else {
                if (rel == Rel.CHILD) {
                    return new VisualNode(prevVisualNode, null, new NoShape());
                } else if (rel == Rel.NEXT) {
                    return new VisualNode(prevVisualNode.getParent(), prevVisualNode, new NoShape());
                } else {
                    return new VisualNode(null, null, new NoShape());
                }
            }
        }


        /**
         * Creates children for the specified visual node that corresponds to the specified schema node and
         * updates the size of the children subtree of the specified node.
         *
         * @param curSchemaNode       schema node
         * @param node                visual node for the schema node, for which the children are created
         * @param nodeDim             dimensions for the visual node
         * @param offset              the frame offset of node inside its children subtree
         * @param childrenSubtreeSize the size of the children subtree
         */
        private void createChildren(SchemaNode curSchemaNode, VisualNode node, int[] nodeDim, Size offset, Size childrenSubtreeSize) {
            Iterator<java.util.List<String>> conditions = null;
            boolean hasConditions = curSchemaNode instanceof BranchNode;
            if (hasConditions) {
                conditions = ((BranchNode) curSchemaNode).getConditions();
            }
            // maximum height of the top part of the children branch-shapes
            int maxChildTop = 0;
            for (Iterator<SchemaNode> i = curSchemaNode.getChildren(); i.hasNext();) {
                SchemaNode schemaChild = i.next();

                // create the children shape-only, no-schema nodes
                String label = "";
                if (hasConditions) {
                    for (String s : conditions.next()) {
                        label += s + String.format("%n");
                    }
                }

                VisualNode child;
                if (i.hasNext()) {
                    child = new VisualNode(node, null, new ContinuedBranchShape(label));
                } else {
                    child = new VisualNode(node, null, new LastBranchShape(label));
                }
                ((BranchShape) (child.getShape())).setBottomSize(nodeDim[D] + YGAP);
                node.addChild(child);

                // calculate subtree size for the child
                int[] childDim = child.getShape().getShapeDims(graphics);
                calcChildSubtreeSize(child, childDim, schemaChild);

                // add size of the newly created child
                // update maximum top part
                if (maxChildTop < childDim[U]) {
                    maxChildTop = childDim[U];
                }
                // augment to the right
                childrenSubtreeSize.x += child.getSubtreeSize().x;
                int delta;
                if ((delta = child.getSubtreeSize().y - childrenSubtreeSize.y) > 0) {
                    childrenSubtreeSize.y += delta;
                }
            } // loop for schema node children

            // update node offset to fit the heights of the children or the chidren subtree height
            //  if the node itself is higher than any of the children
            // here higher means top part has larger height
            int delta;
            if ((delta = maxChildTop - (nodeDim[U] + offset.y)) > 0) {
                offset.y += delta;
            } else if (node.hasChildren()) {
                // if we have to add the top size of the node to its children subtree size
                childrenSubtreeSize.y += (-delta);
            }
            childrenSubtreeSize.y += node.getClosingShape().getShapeDims(graphics)[U];
        }

        /**
         * Calculates size of the child node's shape-node subtree.
         *
         * @param child       the child shape-node
         * @param childDim    dimensions of the child shape-node
         * @param schemaChild the child in the schema tree of the node that corresponds to the child of this node in the visual tree
         */
        private void calcChildSubtreeSize(VisualNode child, int[] childDim, SchemaNode schemaChild) {
            int leftSubtreeSize = childDim[L] + XGAP;
            int rightSubtreeSize = childDim[R] + XGAP;
            int subtreeSizeY = childDim[U] + childDim[D] + YGAP;

            // get the size of the subtree of the child's next node
            VisualNode childNext = visit(schemaChild, child, Rel.NEXT);
            // augment the size of the child's node to contain its subtree also
            int delta;
            if ((delta = childNext.getLeftSubtreeSize() - leftSubtreeSize) > 0) {
                leftSubtreeSize += delta;
            }
            if ((delta = childNext.getRightSubtreeSize() - rightSubtreeSize) > 0) {
                rightSubtreeSize += delta;
            }
            subtreeSizeY += childNext.getSubtreeSize().y;
            child.setSubtreeSize(leftSubtreeSize, rightSubtreeSize, subtreeSizeY);
        }
    }

    /**
     * Builds a visual tree basing on the specified schema tree and returns the root of the created tree.
     *
     * @param schemaTree the tree upon which the visual tree is built
     * @param g          graphics to take visual parameters (eg., font metrics) from
     * @return the root of the visual tree
     */
    public VisualNode build(SchemaTree schemaTree, Graphics g) {
        return (new VisualTreeBuilderImpl(schemaTree, g)).build();
    }
}
