package ru.amse.baltijsky.javascheme.exporter.stream;

import ru.amse.baltijsky.javascheme.tree.BranchNode;
import ru.amse.baltijsky.javascheme.tree.NodeType;
import static ru.amse.baltijsky.javascheme.tree.NodeType.*;
import ru.amse.baltijsky.javascheme.tree.SchemaNode;
import ru.amse.baltijsky.javascheme.tree.walker.ITreeWalker;
import ru.amse.baltijsky.javascheme.tree.walker.TraversalState;

import java.io.PrintWriter;
import java.util.*;

/**
 * Exports the tree to the given writer in DOT language. Then it may be visualized via GraphViz or similar tools.
 * Objects of this class are reentrant (ie. the <code>export</code> method may be called several times).
 * It is necessary to make the output file more or less human-readable; to afford it, we'll separate
 * node and edge descriptions.
 * Nodes have unique names starting with its type and followed by a specific ID number following an underscore "_",
 * eg. <code>while_1</code>.
 * Each auxiliary node that relates to a certain tree node starts with the full name of the tree node followed by
 * the prefix specifying the type of the auxiliary node, eg. <code>while_1_tail_1</code>.
 * Also, to achieve the necessary graph layout, node groupings are created. They correspond to the same nested level
 * and their identifiers are of the form <code>g_[level]_[number]</code>, where <code>level</code>
 * is the nesting level of the nodes of the group and <code>number</code> is the number of the group of the current
 * nesting level (they are graph-global). eg. <code>g_1_3</code> stands for the group of nodes that are on the second
 * level, and there already were 3 nesting consrtuctions of the same level.
 */
public class DotExporter implements IStreamExporter {
    /**
     * The implementation of DOT exporter. Non-reentrant, its objects are created each time the <code>export</code>
     * method is called by the user.
     */
    private class DotExporterImpl {
        /**
         * Provides facilities for naming nodes and node groups.
         * Also contains current ID numbers for different types of nodes.
         */
        private class Namer {
            /**
             * ID numbers array for each type of the node.
             */
            private int[] ids = new int[NodeType.NODE_TYPES_AMOUNT];
            /**
             * List of current group ID numbers for all levels of the tree encountered by now. As new levels
             * are reached, their current group IDs are appended to the list.
             */
            private List<Integer> groupIds = new ArrayList<Integer>();

            /**
             * Returns a new ID number for the node of the specified type.
             */
            public int getNewIdNumber(NodeType type) {
                return ids[type.toInt()]++;
            }

            /**
             * Returns a new group ID and appends the ID list if necessary.
             */
            public String createGroupId() {
                while (groupIds.size() - 1 < state.depth) {
                    groupIds.add(0);
                }
                int newId = groupIds.get(state.depth) + 1;
                groupIds.set(state.depth, newId);

                return "g_" + state.depth + "_" + newId;
            }

            public String getCurrentGroupId() {
                if (groupIds.size() - 1 < state.depth) {
                    return createGroupId();
                }
                return "g_" + state.depth + "_" + groupIds.get(state.depth);
            }

            public String getElderGroupId() {
                if (state.depth == 0) {
                    return "";
                }
                if (groupIds.size() - 1 < state.depth - 1) {
                    return createGroupId();
                }
                return "g_" + (state.depth - 1) + "_" + groupIds.get(state.depth - 1);
            }
        }

        /**
         * Provides a convenient way for setting edge attributes.
         */
        private class EdgeAttr {
            /**
             * The edge is invisible (used to impose node ordering).
             */
            public static final int INVIS = 1 << 0;
            /**
             * The edge has no arrow (used in complex transition, such as in <code>while</code>).
             */
            public static final int NO_ARROWHEAD = 1 << 1;
            /**
             * The edge is horizontal and directed from left to right.
             */
            public static final int HORIZ_LR = 1 << 2;
            /**
             * The edge is horizontal and directed from right to left.
             */
            public static final int HORIZ_RL = 1 << 3;
            /**
             * The edge is vertical and goes bottom up.
             */
            public static final int VERT_INV = 1 << 4;
            /**
             * The edge is vertical, goes bottom up and meets head node in its southeastern part.
             */
            public static final int VERT_SE = 1 << 5;
            /**
             * The edge has eastern headport.
             */
            public static final int HEAD_E = 1 << 6;
            /**
             * The edge is not expected to impose rank ordering on edges.
             */
            public static final int NO_CONSTRAINT = 1 << 7;
            /**
             * Placeholder: insert new edge styles before it and fix the value to be the greatest.
             */
            public static final int LAST = 1 << 8;

            /**
             * Writes out a transition (edge of the <code>DOT</code> graph).
             * Used to create single-line attributes assignment.
             *
             * @param from edge tail node
             * @param to   edge head node
             * @return <code>this</code> to allow call chain
             */
            public EdgeAttr writeEdge(String from, String to) {
                prn.writeIndented(from + " -> " + to + " ");
                return this;
            }

            /**
             * Writes style attributes' definitions for the edge.
             *
             * @param style the desired style (bit-setting)
             * @return <code>this</code> to allow chain call
             */
            public EdgeAttr writeAttributes(int style) {
                StringBuilder sOut = new StringBuilder("[");
                if ((style & INVIS) != 0) {
                    sOut.append("style = invis, ");
                }
                if ((style & NO_ARROWHEAD) != 0) {
                    sOut.append("arrowhead = none, ");
                }
                if ((style & HORIZ_LR) != 0) {
                    sOut.append("tailport = e, headport = w, ");
                }
                if ((style & HORIZ_RL) != 0) {
                    sOut.append("tailport = w, headport = e, ");
                }
                if ((style & VERT_INV) != 0) {
                    sOut.append("tailport = n, headport = s, ");
                }
                if ((style & VERT_SE) != 0) {
                    sOut.append("tailport = n, headport = se, ");
                }
                if ((style & HEAD_E) != 0) {
                    sOut.append("headport = e, ");
                }
                if ((style & NO_CONSTRAINT) != 0) {
                    sOut.append("constraint = false, ");
                }

                // if we've added at least one attribute description, we need to get rid of the trailing ", "
                if (style != 0 && style < LAST) {
                    sOut.delete(sOut.length() - 2, sOut.length() - 1);
                }
                sOut.append("] ");
                prn.write(sOut.toString());
                return this;
            }

            /**
             * Writes the specific label for the edge.
             *
             * @return <code>this</code> to allow chain call
             */
            public EdgeAttr writeLabel(String label) {
                writeFeature("label", label);
                return this;
            }

            /**
             * Writes the headport for the edge.
             */
            public EdgeAttr writeHeadport(String port) {
                writeFeature("headport", port);
                return this;
            }

            /**
             * Writes the headport for the edge.
             */
            public EdgeAttr writeTailport(String port) {
                writeFeature("tailport", port);
                return this;
            }

            /**
             * Writes out the human-reader convenient semicolon and terminates the line.
             */
            public void writeLineEnd() {
                prn.writeln(";");
            }

            private void writeFeature(String feature, String value) {
                prn.write("[" + feature + " = \"" + value + "\"] ");
            }
        }

        /**
         * Provides a convenient way for setting node attributes.
         */
        private class NodeAttr {
            /**
             * Writes out node name and allows chain call to write the attributes.
             *
             * @return <code>this</code> to allow chain call
             */
            private NodeAttr writeNode(String s) {
                prn.writeIndented(s + " ");
                return this;
            }

            /**
             * Writes out the attributes that are specific for the given node type.
             * The node will be labeled via the given <code>String</code>.
             *
             * @param type  node type
             * @param label the text to write on the node
             * @return <code>this</code> to allow chain call
             */
            public NodeAttr writeAttributes(NodeType type, String label, String group) {
                // currently, all nodes different from considered are handled as Actions.
                switch (type) {
                case ACTION:
                default:
                    writeLabel(label).writeShape("box").writeGroup(group);
                    break;
                case IF:
                    writeLabel("if").writeShape("diamond").writeGroup(group);
                    break;
                case WHILE:
                    writeLabel("while").writeShape("hexagon").writeGroup(group);
                    break;
                }
                return this;
            }

            /**
             * Writes out the attributes that are specific for the given node type with empty label.
             *
             * @param type node type
             * @return <code>this</code> to allow chain call
             */
            public NodeAttr writeAttributes(NodeType type, String group) {
                writeAttributes(type, "", group);
                return this;
            }

            /**
             * Writes out shape attribute.
             *
             * @param shape the shape to assign to the current node
             * @return <code>this</code> to allow chain calls
             */
            public NodeAttr writeShape(String shape) {
                prn.write("[shape = " + shape + "] ");
                return this;
            }

            /**
             * Writes out the label for the current node.
             *
             * @param label the label string
             * @return <code>this</code> to allow chain calls
             */
            public NodeAttr writeLabel(String label) {
                prn.write("[label = \"" + label + "\"] ");
                return this;
            }

            /**
             * Creates and writes out group ID for the current node. We may put the node into the current group
             * or the group one level shallower than the current, depending on the parameter.
             *
             * @param writeCurrentGroup if true, assigns the node current level group; otherwise, it is put into the group
             *                          one level shallower
             * @return <code>this</code> to allow chain calls
             */
            public NodeAttr writeGroup(boolean writeCurrentGroup) {
                prn.write("[group = " + (writeCurrentGroup ? namer.getCurrentGroupId() : namer.getElderGroupId()) + "] ");
                return this;
            }

            /**
             * Writes the specified group identifier as node group.
             *
             * @return <code>this</code> to allow chain calls
             */
            public NodeAttr writeGroup(String group) {
                prn.write("[group = " + group + "] ");
                return this;
            }

            /**
             * Writes out node height.
             *
             * @return <code>this</code> to allow chain calls
             */
            public NodeAttr writeHeight(String height) {
                prn.write("[height = " + height + "] ");
                return this;
            }

            /**
             * Writes out attributes for an invisible auxiliary node. It may be put into the current group or into the
             * group one level shallower, depending on the parameter.
             *
             * @param writeCurrentGroup if true, assigns the node current level group; otherwise, it is put into the group
             *                          one level shallower
             * @return <code>this</code> to allow chain calls
             */
            public NodeAttr writeAuxNodeAttributes(boolean writeCurrentGroup) {
                prn.write("[style = invis, height = 0, width = 0, fixedsize] ");
                writeLabel("");
                writeGroup(writeCurrentGroup);
                return this;
            }

            /**
             * Writes out the human-reader convenient semicolon and terminates the line.
             */
            public void writeLineEnd() {
                prn.writeln(";");
            }

        }

        /**
         * Works with the output providing prettyprint.
         */
        private class PrettyPrinter {
            /**
             * Indentation string in the generated DOT file.
             */
            private static final String INDENT = "    ";
            /**
             * The writer.
             */
            private PrintWriter out;
            /**
             * Nesting level in the output DOT file (used for prettyprint).
             */
            private int printLevel;

            /**
             * Constructs a new pretty printer for the specified output printer.
             */
            public PrettyPrinter(PrintWriter newOut) {
                out = newOut;
            }

            /**
             * Increases indentation.
             */
            public void incIndent() {
                ++printLevel;
            }

            /**
             * Decreases indentation.
             */
            public void decIndent() {
                --printLevel;
            }

            /**
             * Writes a string to the output indenting it by <code>printLevel</code> positions, terminates the line.
             *
             * @param s the string to write
             */
            public void writelnIndented(String s) {
                writeIndent();
                writeln(s);
            }

            /**
             * Writes a string to the output without indentation, terminates the line.
             *
             * @param s the string to write
             */
            public void writeln(String s) {
                out.println(s);
            }

            /**
             * Writes a blank line to the output.
             */
            public void writeln() {
                out.println();
            }

            /**
             * Writes a string not terminating the line, indenting the string by <code>printLevel</code> positions.
             *
             * @param s the dtring to write
             */
            public void writeIndented(String s) {
                writeIndent();
                write(s);
            }

            /**
             * Writes a string not terminating the line, without indentation.
             *
             * @param s the dtring to write
             */
            public void write(String s) {
                out.print(s);
            }

            /**
             * Writes leading blanks to the output to provide indentation.
             */
            private void writeIndent() {
                for (int i = 0; i < printLevel; ++i) {
                    out.print(INDENT);
                }
            }
        }

        /**
         * The tree walker.
         */
        private ITreeWalker<SchemaNode> treeWalker;
        /**
         * Current traversal state.
         */
        private TraversalState<SchemaNode> state;
        /**
         * ID counter.
         */
        private Namer namer = new Namer();
        /**
         * Edge attribute writer.
         */
        private EdgeAttr edgeAttr = new EdgeAttr();
        /**
         * Node attribute writer.
         */
        private NodeAttr nodeAttr = new NodeAttr();
        /**
         * Prettyprinter.
         */
        private PrettyPrinter prn;
        /**
         * A map from schema tree nodes to <code>DOT</code> graph nodes.
         * We'll create nodes in map first, write out the transitions, and then write out the nodes
         * (we need the order, so it's LinkedHashMap)
         */
        private Map<SchemaNode, String> nodeNameMap = new LinkedHashMap<SchemaNode, String>();
        /**
         * A map from nodes to their group identifiers.
         * Used to set proper node groupings when writing nodes to the <code>DOT</code> file.
         */
        private Map<SchemaNode, String> nodeGroupMap = new LinkedHashMap<SchemaNode, String>();
        /**
         * Conditions of the branches for all currently existing schema tree levels.
         */
        private Deque<Iterator<List<String>>> branchConditions = new LinkedList<Iterator<List<String>>>();

        /**
         * Constructor.
         *
         * @param newOut        the printer
         * @param newTreeWalker the tree walker
         */
        public DotExporterImpl(PrintWriter newOut, ITreeWalker<SchemaNode> newTreeWalker) {
            treeWalker = newTreeWalker;
            treeWalker.reset();
            prn = new PrettyPrinter(newOut);
        }

        /**
         * Exports the whole tree.
         */
        public void export() {
            writeGraphHeader();
            if (treeWalker.hasNextState()) {
                state = treeWalker.getNextState();
                writeClassHeader();
                writeMethodHeader();
                writeInitialAndFinalState();
                writeTransitions();
                writeNodes();
                writeMethodEnd();
                writeClassEnd();
            }
            writeGraphEnd();
        }

        /**
         * Writes graph header to the output file.
         * The header consists of graph declaration and common settings for graph components.
         */
        private void writeGraphHeader() {
            prn.writelnIndented("digraph ProgramFlow");
            prn.writelnIndented("{");
            prn.incIndent();
            prn.writelnIndented("edge [weight = 100000, headport = n, tailport = s, constraint = true, arrowsize = 0.7, fontname = Courier, labeldistance = 1.5];");
            prn.writelnIndented("size = \"100,100\";");
            prn.writelnIndented("node [fontname = Courier];");
            prn.writelnIndented("fontname = Courier;");
        }

        /**
         * Writes graph end.
         */
        private void writeGraphEnd() {
            prn.decIndent();
            prn.writelnIndented("}");
        }

        /**
         * Writes class claster header.
         */
        private void writeClassHeader() {
            skipToNode(CLASS);

            prn.writelnIndented("subgraph clusterClass");
            prn.writelnIndented("{");
            prn.incIndent();
            if (state.node.getNodeType() == CLASS) { // the tree's not empty
                prn.writelnIndented("label = \"class\\n" + state.node.getCodeSingleLine() + "\";");
            }
        }

        private void writeClassEnd() {
            prn.decIndent();
            prn.writelnIndented("}");
        }

        private void writeMethodHeader() {
            skipToNode(METHOD);
            prn.writelnIndented("subgraph clusterFunc_" + namer.getNewIdNumber(METHOD));
            prn.writelnIndented("{");
            prn.incIndent();
            prn.writelnIndented("label = \"" + state.node.getCodeSingleLine() + "\";");
            prn.writelnIndented("ordering = out;");
        }

        private void writeMethodEnd() {
            prn.decIndent();
            prn.writelnIndented("}");
        }

        private void writeInitialAndFinalState() {
            prn.writelnIndented("// draw initial state in the automata-like style");
            prn.writelnIndented("{");
            prn.incIndent();
            prn.writelnIndented("rank = same;");
            nodeAttr.writeNode("preinit").writeAuxNodeAttributes(false).writeLineEnd();
            nodeAttr.writeNode("initial").writeShape("circle").writeLabel("").writeGroup(true).writeLineEnd();
            prn.decIndent();
            prn.writelnIndented("}");
            prn.writelnIndented("");
            edgeAttr.writeEdge("preinit", "initial").writeAttributes(EdgeAttr.HORIZ_LR).writeLineEnd();
            prn.writelnIndented("");
            prn.writelnIndented("// final state");
            nodeAttr.writeNode("final").writeLabel("").writeHeight("0.35").writeShape("doublecircle").writeGroup(true).writeLineEnd();
        }

        private void writeTransitions() {
            prn.writeln();
            prn.writelnIndented("// transitions");
            prn.writelnIndented("////////////////////////////////////////////////////////////////////////////////");
            // for handling the initial transition case (it has a special previous node)
            boolean wroteInitial = false;
            while (treeWalker.hasNextState()) {
                TraversalState<SchemaNode> lastState = state;
                state = treeWalker.getNextState();

                // close nesting levels if necessary
                closeNestingLevels(lastState);

                // add node to the node map
                String newNodeName = createNode();

                // add transition
                if (!wroteInitial) {
                    edgeAttr.writeEdge("initial", newNodeName).writeLineEnd();
                    wroteInitial = true;
                } else {
                    writeTransition(newNodeName);
                }
            }

            // add the transition from the last node to the final node
            if (state != null) {
                edgeAttr.writeEdge(nodeNameMap.get(state.node), "final").writeLineEnd();
            }
        }

        /**
         * Creates a node and pushes it into the node map to be written out later.
         *
         * @return name of the new node
         */
        private String createNode() {
            String newNodeName = state.node.getNodeType().getDescription() + "_" + namer.getNewIdNumber(state.node.getNodeType());
            nodeNameMap.put(state.node, newNodeName);

            // put the node in the appropriate group
            String groupId;
            if (state.rel == TraversalState.Rel.CHILD) {
                groupId = namer.createGroupId();
            } else {
                groupId = namer.getCurrentGroupId();
            }
            nodeGroupMap.put(state.node, groupId);

            // if the node is a branching node, save the iterator over its branch conditions
            if (state.node instanceof BranchNode) {
                branchConditions.addFirst(((BranchNode) state.node).getConditions());
            }

            return newNodeName;
        }

        /**
         * Writes out a new transition.
         *
         * @param newNodeName the name of the node which is the endpoint of the new transition
         */
        private void writeTransition(String newNodeName) {
            edgeAttr.writeEdge(nodeNameMap.get(state.prev), newNodeName);
            if (state.rel == TraversalState.Rel.NEXT) {
                // ok
            } else if (state.rel == TraversalState.Rel.CHILD) {
                String label;
                if (state.node instanceof BranchNode) {
                    label = Arrays.toString(branchConditions.peekFirst().next().toArray());
                    if (!branchConditions.peekFirst().hasNext()) {
                        branchConditions.removeFirst();
                    }
                } else {
                    label = state.prev.getCodeSingleLine();
                }
                edgeAttr.writeTailport("e").writeLabel(label);
            }
            edgeAttr.writeLineEnd();
        }

        /**
         * If the new node is outside the recently considered nesting level, closes the level in the appropriate way
         * (adds ending transitions to loops, if branches etc).
         *
         * @param lastState
         */
        private void closeNestingLevels(TraversalState<SchemaNode> lastState) {
            SchemaNode cur = lastState.node;
            for (int l = lastState.depth; l > state.depth; --l) {
                SchemaNode parent = cur.getParent();
                switch (parent.getNodeType()) {
                case IF:
                    processIfClose(cur, parent);
                    break;
                case WHILE:
                    processWhileClose(cur, parent);
                    break;
                }
                cur = parent;
            }
        }

        private void processIfClose(SchemaNode cur, SchemaNode parent) {
            String parentNodeName = nodeNameMap.get(parent);
            String auxNodeName = parentNodeName + "_tail";
            edgeAttr.writeEdge(nodeNameMap.get(cur), auxNodeName).writeAttributes(EdgeAttr.HEAD_E | EdgeAttr.NO_ARROWHEAD).writeLineEnd();
            edgeAttr.writeEdge(parentNodeName, auxNodeName).writeAttributes(EdgeAttr.NO_ARROWHEAD).writeLineEnd();
            // substitute auxiliary edge in the place of the if node parent
            nodeNameMap.put(parent, auxNodeName);
        }

        private void processWhileClose(SchemaNode last, SchemaNode parent) {
            String parentNodeName = nodeNameMap.get(parent);
            String auxNodeName = parentNodeName + "_tail";
            edgeAttr.writeEdge(nodeNameMap.get(last), auxNodeName).writeAttributes(EdgeAttr.HEAD_E | EdgeAttr.NO_ARROWHEAD).writeLineEnd();
            edgeAttr.writeEdge(parentNodeName, auxNodeName).writeAttributes(EdgeAttr.NO_ARROWHEAD).writeLineEnd();
            // substitute auxiliary edge in the place of the if node parent
            nodeNameMap.put(parent, auxNodeName);
        }

        private void writeNodes() {
            prn.writeln();
            prn.writelnIndented("// nodes");
            prn.writelnIndented("////////////////////////////////////////////////////////////////////////////////");
            for (Map.Entry<SchemaNode, String> e : nodeNameMap.entrySet()) {
                String nodeName = e.getValue();
                if (nodeName.equals("initial")) {
                    continue;
                }
                SchemaNode p = e.getKey();

                // get rid of the auxiliary suffix
                int suffixIndex = nodeName.indexOf("_", nodeName.indexOf("_") + 1);
                if (suffixIndex > 0) {
                    nodeName = nodeName.substring(0, suffixIndex);
                }
                String nodeGroup = nodeGroupMap.get(p);
                nodeAttr.writeNode(nodeName).writeAttributes(p.getNodeType(), p.getCodeSingleLine(), nodeGroup).writeLineEnd();
                // add auxiilary nodes if necessary
                if (p.getNodeType() == IF || p.getNodeType() == WHILE) {
                    nodeAttr.writeNode(nodeName + "_tail").writeAuxNodeAttributes(true).writeLineEnd();
                }
            }
        }

        ///////////////////////////////////// auxiliary methods

        /**
         * Skips through the internal tree nodes until a node of the specified type is encountered.
         *
         * @param type the type of the node to skip to.
         */
        private void skipToNode(NodeType type) {
            while (treeWalker.hasNextState() && state.node.getNodeType() != type) {
                state = treeWalker.getNextState();
            }
        }

    }

    /**
     * @see IStreamExporter
     */
    public void export(PrintWriter out, ITreeWalker<SchemaNode> treeWalker) {
        (new DotExporterImpl(out, treeWalker)).export();
    }
}
